
# `A exp(-a s^2) + B exp(-b s^2) + C exp(-c s^2) + D exp(-d s^2) + E`
struct ExpandedBesselIntegral
    A :: Float64
    a :: Float64
    B :: Float64
    b :: Float64
    C :: Float64
    c :: Float64
    D :: Float64
    d :: Float64
    E :: Float64
end

struct FormFactor
    j0 :: ExpandedBesselIntegral
    j2 :: ExpandedBesselIntegral
    config :: String
    g :: Float64
end

const identity_form_factor = let
    j0 = ExpandedBesselIntegral(0, 0, 0, 0, 0, 0, 0, 0, 1)
    j2 = ExpandedBesselIntegral(0, 0, 0, 0, 0, 0, 0, 0, 0)
    g = 2
    FormFactor(j0, j2, "", g)
end

"""
    FormFactor(ion::String; g_lande=2)

The magnetic form factor for a given magnetic ion and charge state. When passed
to an [`intensity_formula`](@ref), it rescales structure factor intensities
based on the magnitude of the scattering vector, ``|𝐪|``.

The parameter `ion` must be one of the following strings:

```
Am2, Am3, Am4, Am5, Am6, Am7, Au1, Au2, Au3, Au4, Au5, Ce2, Co0, Co1, Co2, Co3,
Co4, Cr0, Cr1, Cr2, Cr3, Cr4, Cu0, Cu1, Cu2, Cu3, Cu4, Dy2, Dy3, Er2, Er3, Eu2,
Eu3, Fe0, Fe1, Fe2, Fe3, Fe4, Gd2, Gd3, Hf2, Hf3, Ho2, Ho3, Ir0a, Ir0b, Ir0c,
Ir1a, Ir1b, Ir2, Ir3, Ir4, Ir5, Ir6, Mn0, Mn1, Mn2, Mn3, Mn4, Mn5, Mo0, Mo1, Nb0,
Nb1, Nd2, Nd3, Ni0, Ni1, Ni2, Ni3, Ni4, Np3, Np4, Np5, Np6, Os0a, Os0b, Os0c,
Os1a, Os1b, Os2, Os3, Os4, Os5, Os6, Os7, Pd0, Pd1, Pr3, Pt1, Pt2, Pt3, Pt4,
Pt5, Pt6, Pu3, Pu4, Pu5, Pu6, Re0a, Re0b, Re0c, Re1a, Re1b, Re2, Re3, Re4, Re5,
Re6, Rh0, Rh1, Ru0, Ru1, Sc0, Sc1, Sc2, Sm2, Sm3, Ta2, Ta3, Ta4, Tb2, Tb3, Tc0,
Tc1, Ti0, Ti1, Ti2, Ti3, Tm2, Tm3, U3, U4, U5, V0, V1, V2, V3, V4, W0a, W0b,
W0c, W1a, W1b, W2c, W3, W4, W5, Y0, Yb2, Yb3, Zr0, Zr1
```

The trailing number denotes ionization state. For example, `"Fe0"` denotes a
neutral iron atom, while `"Fe2"` denotes `Fe²⁺`. If multiple electronic
configurations are possible, they will be distinguished by a trailing letter
(`a`, `b`, ...). Omitting this letter will print an informative error,

```
FormFactor("Ir0")

ERROR: Disambiguate form factor according to electronic configuration:
    "Ir0a" -- 6s⁰5d⁹
    "Ir0b" -- 6s¹5d⁸
    "Ir0c" -- 6s²5d⁷
```

In the dipolar approximation (small ``|𝐪|``) the form factor is

``F(s) = ⟨j_0(s)⟩ + \\frac{2-g}{g} ⟨j_2(s)⟩``,

involving ``s = |𝐪|/4π`` and the Landé ``g``-factor. The ``⟨j_l(s)⟩`` are
radial averages of the ``l``th spherical Bessel function of the magnetic dipole.
More details are provided in Ref. [1].

The ``⟨j_l(s)⟩`` can be approximated as a sum of Gaussians,

```math
⟨j_0(s)⟩ = A e^{-as^2} + B e^{-bs^2} + C e^{-cs^2} + D e^{-ds^2} + E \\\\
⟨j_2(s)⟩ = (A e^{-as^2} + B e^{-bs^2} + C e^{-cs^2} + D e^{-ds^2} + E) s^2
```

For 3d, 4d, rare earth, and actinide ions, Sunny uses the revised tables of P.
J. Brown, as documented in the McPhase package [2]. For 5d ions, Sunny uses the
tables of Kobayashi, Nagao, Ito [3].

References:

 1. [P. J. Brown, The Neutron Data Booklet, 2nd ed., Sec. 2.5 Magnetic Form
    Factors (2003)](https://www.ill.eu/sites/ccsl/ffacts/ffachtml.html)
 2. Coefficient tables in [McPhase
    documentation](https://www2.cpfs.mpg.de/~rotter/homepage_mcphase/manual/node137.html)
 3. K. Kobayashi, T. Nagao, M. Ito, Acta Cryst. A, 67 pp 473–480 (2011)
"""
function FormFactor(ion::String; g_lande=2)
    if !haskey(radial_integral_coefficients, ion)
        if !haskey(radial_integral_coefficients, ion*"a")
            error("Form factor requires species name and charge state, e.g. \"Fe2\" for Fe²⁺")
        else
            avail_keys = [k for k in keys(radial_integral_coefficients) if startswith(k, ion)]
            avail_strs = map(sort(avail_keys)) do k
                "    $(repr(k)) -- " * radial_integral_coefficients[k][3]
            end
            error("""
                Disambiguate form factor according to electronic configuration:
                $(join(avail_strs, "\n"))
                """)
        end
    end

    (j0, j2, config) = radial_integral_coefficients[ion]
    j0 = ExpandedBesselIntegral(j0...)
    j2 = ExpandedBesselIntegral(j2...)
    FormFactor(j0, j2, config, g_lande)
end


function compute_gaussian_expansion(j::ExpandedBesselIntegral, s2)
    (; A, a, B, b, C, c, D, d, E) = j
    return A*exp(-a*s2) + B*exp(-b*s2) + C*exp(-c*s2) + D*exp(-d*s2) + E
end

function compute_form_factor(form_factor::FormFactor, k2_absolute::Float64)
    (; j0, j2, g) = form_factor

    # Return early if this is the identity form factor
    (j0.A == j0.B == j0.C == j0.D == 0) && (j0.E == 1) && (g == 2) && return 1.0

    s2 = k2_absolute / (4π)^2
    if g == 2
        return compute_gaussian_expansion(j0, s2)
    else
        J0 = compute_gaussian_expansion(j0, s2)
        J2 = compute_gaussian_expansion(j2, s2) * s2
        return J0 + ((2-g)/g) * J2
    end
end

# Given a form factor for each "symmetry class" of sites, return a form factor
# for each atom in the crystal.
function propagate_form_factors_to_atoms(ffs, cryst::Crystal)
    isnothing(ffs) && return fill(identity_form_factor, natoms(cryst))
    
    ref_classes = unique(cryst.classes)
    if length(ffs) != length(ref_classes)
        error("""Received $(length(ffs)) form factors, but $(length(ref_classes)) are
                 required, one for each symmetry-distinct site in the crystal.""")
    end

    return [ffs[findfirst(==(c), ref_classes)] for c in cryst.classes]
end


const radial_integral_coefficients = Dict(
    # 3d electrons of transition atoms and ions (P. J. Brown)
    "Sc0" => ([0.2512, 90.0296, 0.3290, 39.4021, 0.4235, 14.3222, 0, 0, -0.0043], [10.8172, 54.3270, 4.7353, 14.8471, 0.6071, 4.2180, 0, 0, 0.0011], ""),
    "Sc1" => ([0.4889, 51.1603, 0.5203, 14.0764, -0.0286, 0.1792, 0, 0, 0.0185],  [8.5021, 34.2851, 3.2116, 10.9940, 0.4244, 3.6055, 0, 0, 0.0009], ""),
    "Sc2" => ([0.5048, 31.4035, 0.5186, 10.9897, -0.0241, 1.1831, 0, 0, 0.0000],  [4.3683, 28.6544, 3.7231, 10.8233, 0.6074, 3.6678, 0, 0, 0.0014], ""),
    "Ti0" => ([0.4657, 33.5898, 0.5490, 9.8791, -0.0291, 0.3232, 0, 0, 0.0123],   [4.3583, 36.0556, 3.8230, 11.1328, 0.6855, 3.4692, 0, 0, 0.0020], ""),
    "Ti1" => ([0.5093, 36.7033, 0.5032, 10.3713, -0.0263, 0.3106, 0, 0, 0.0116],  [6.1567, 27.2754, 2.6833, 8.9827, 0.4070, 3.0524, 0, 0, 0.0011], ""),
    "Ti2" => ([0.5091, 24.9763, 0.5162, 8.7569, -0.0281, 0.9160, 0, 0, 0.0015],   [4.3107, 18.3484, 2.0960, 6.7970, 0.2984, 2.5476, 0, 0, 0.0007], ""),
    "Ti3" => ([0.3571, 22.8413, 0.6688, 8.9306, -0.0354, 0.4833, 0, 0, 0.0099],   [3.3717, 14.4441, 1.8258, 5.7126, 0.2470, 2.2654, 0, 0, 0.0005], ""),
    "V0" =>  ([0.4086, 28.8109, 0.6077, 8.5437, -0.0295, 0.2768, 0, 0, 0.0123],   [3.7600, 21.8313, 2.4026, 7.5458, 0.4464, 2.6628, 0, 0, 0.0017], ""),
    "V1" =>  ([0.4444, 32.6479, 0.5683, 9.0971, -0.2285, 0.0218, 0, 0, 0.2150],   [4.7474, 23.3226, 2.3609, 7.8082, 0.4105, 2.7063, 0, 0, 0.0014], ""),
    "V2" =>  ([0.4085, 23.8526, 0.6091, 8.2456, -0.1676, 0.0415, 0, 0, 0.1496],   [3.4386, 16.5303, 1.9638, 6.1415, 0.2997, 2.2669, 0, 0, 0.0009], ""),
    "V3" =>  ([0.3598, 19.3364, 0.6632, 7.6172, -0.3064, 0.0296, 0, 0, 0.2835],   [2.3005, 14.6821, 2.0364, 6.1304, 0.4099, 2.3815, 0, 0, 0.0014], ""),
    "V4" =>  ([0.3106, 16.8160, 0.7198, 7.0487, -0.0521, 0.3020, 0, 0, 0.0221],   [1.8377, 12.2668, 1.8247, 5.4578, 0.3979, 2.2483, 0, 0, 0.0012], ""),
    "Cr0" => ([0.1135, 45.1990, 0.3481, 19.4931, 0.5477, 7.3542, 0, 0, -0.0092],  [3.4085, 20.1267, 2.1006, 6.8020, 0.4266, 2.3941, 0, 0, 0.0019], ""),
    "Cr1" => ([-0.0977, 0.0470, 0.4544, 26.0054, 0.5579, 7.4892, 0, 0, 0.0831],   [3.7768, 20.3456, 2.1028, 6.8926, 0.4010, 2.4114, 0, 0, 0.0017], ""),
    "Cr2" => ([1.2024, -0.0055, 0.4158, 20.5475, 0.6032, 6.9560, 0, 0, -1.2218],  [2.6422, 16.0598, 1.9198, 6.2531, 0.4446, 2.3715, 0, 0, 0.0020], ""),
    "Cr3" => ([-0.3094, 0.0274, 0.3680, 17.0355, 0.6559, 6.5236, 0, 0, 0.2856],   [1.6262, 15.0656, 2.0618, 6.2842, 0.5281, 2.3680, 0, 0, 0.0023], ""),
    "Cr4" => ([-0.2320, 0.0433, 0.3101, 14.9518, 0.7182, 6.1726, 0, 0, 0.2042],   [1.0293, 13.9498, 1.9933, 6.0593, 0.5974, 2.3457, 0, 0, 0.0027], ""),
    "Mn0" => ([0.2438, 24.9629, 0.1472, 15.6728, 0.6189, 6.5403, 0, 0, -0.0105],  [2.6681, 16.0601, 1.7561, 5.6396, 0.3675, 2.0488, 0, 0, 0.0017], ""),
    "Mn1" => ([-0.0138, 0.4213, 0.4231, 24.6680, 0.5905, 6.6545, 0, 0, -0.0010],  [3.2953, 18.6950, 1.8792, 6.2403, 0.3927, 2.2006, 0, 0, 0.0022], ""),
    "Mn2" => ([0.4220, 17.6840, 0.5948, 6.0050, 0.0043, -0.6090, 0, 0, -0.0219],  [2.0515, 15.5561, 1.8841, 6.0625, 0.4787, 2.2323, 0, 0, 0.0027], ""),
    "Mn3" => ([0.4198, 14.2829, 0.6054, 5.4689, 0.9241, -0.0088, 0, 0, -0.9498],  [1.2427, 14.9966, 1.9567, 6.1181, 0.5732, 2.2577, 0, 0, 0.0031], ""),
    "Mn4" => ([0.3760, 12.5661, 0.6602, 5.1329, -0.0372, 0.5630, 0, 0, 0.0011],   [0.7879, 13.8857, 1.8717, 5.7433, 0.5981, 2.1818, 0, 0, 0.0034], ""),
    "Mn5" => ([0.2924, 11.6655, 0.7405, 5.0741, -1.7883, 0.0059, 0, 0, 1.7557],   [-0.2394, 10.7309, -0.1190, 6.5989, 0.3505, 1.4912, 0, 0, 0.0078], ""),
    "Fe0" => ([0.0706, 35.0085, 0.3589, 15.3583, 0.5819, 5.5606, 0, 0, -0.0114],  [1.9405, 18.4733, 1.9566, 6.3234, 0.5166, 2.1607, 0, 0, 0.0036], ""),
    "Fe1" => ([0.1251, 34.9633, 0.3629, 15.5144, 0.5223, 5.5914, 0, 0, -0.0105],  [2.6290, 18.6598, 1.8704, 6.3313, 0.4690, 2.1628, 0, 0, 0.0031], ""),
    "Fe2" => ([0.0263, 34.9597, 0.3668, 15.9435, 0.6188, 5.5935, 0, 0, -0.0119],  [1.6490, 16.5593, 1.9064, 6.1325, 0.5206, 2.1370, 0, 0, 0.0035], ""),
    "Fe3" => ([0.3972, 13.2442, 0.6295, 4.9034, -0.0314, 0.3496, 0, 0, 0.0044],   [1.3602, 11.9976, 1.5188, 5.0025, 0.4705, 1.9914, 0, 0, 0.0038], ""),
    "Fe4" => ([0.3782, 11.3800, 0.6556, 4.5920, -0.0346, 0.4833, 0, 0, 0.0005],   [1.5582, 8.2750, 1.1863, 3.2794, 0.1366, 1.1068, 0, 0, -0.0022], ""),
    "Co0" => ([0.4139, 16.1616, 0.6013, 4.7805, -0.1518, 0.0210, 0, 0, 0.1345],   [1.9678, 14.1699, 1.4911, 4.9475, 0.3844, 1.7973, 0, 0, 0.0027], ""),
    "Co1" => ([0.0990, 33.1252, 0.3645, 15.1768, 0.5470, 5.0081, 0, 0, -0.0109],  [2.4097, 16.1608, 1.5780, 5.4604, 0.4095, 1.9141, 0, 0, 0.0031], ""),
    "Co2" => ([0.4332, 14.3553, 0.5857, 4.6077, -0.0382, 0.1338, 0, 0, 0.0179],   [1.9049, 11.6444, 1.3159, 4.3574, 0.3146, 1.6453, 0, 0, 0.0017], ""),
    "Co3" => ([0.3902, 12.5078, 0.6324, 4.4574, -0.1500, 0.0343, 0, 0, 0.1272],   [1.7058, 8.8595, 1.1409, 3.3086, 0.1474, 1.0899, 0, 0, -0.0025], ""),
    "Co4" => ([0.3515, 10.7785, 0.6778, 4.2343, -0.0389, 0.2409, 0, 0, 0.0098],   [1.3110, 8.0252, 1.1551, 3.1792, 0.1608, 1.1301, 0, 0, -0.0011], ""),
    "Ni0" => ([-0.0172, 35.7392, 0.3174, 14.2689, 0.7136, 4.5661, 0, 0, -0.0143], [1.0302, 12.2521, 1.4669, 4.7453, 0.4521, 1.7437, 0, 0, 0.0036], ""),
    "Ni1" => ([0.0705, 35.8561, 0.3984, 13.8042, 0.5427, 4.3965, 0, 0, -0.0118],  [2.1040, 14.8655, 1.4302, 5.0714, 0.4031, 1.7784, 0, 0, 0.0034], ""),
    "Ni2" => ([0.0163, 35.8826, 0.3916, 13.2233, 0.6052, 4.3388, 0, 0, -0.0133],  [1.7080, 11.0160, 1.2147, 4.1031, 0.3150, 1.5334, 0, 0, 0.0018], ""),
    "Ni3" => ([0.0012, 34.9998, 0.3468, 11.9874, 0.6667, 4.2518, 0, 0, -0.0148],  [1.4683, 8.6713, 1.1068, 3.2574, 0.1794, 1.1058, 0, 0,-0.0023], ""),
    "Ni4" => ([-0.0090, 35.8614, 0.2776, 11.7904, 0.7474, 4.2011, 0, 0, -0.0163], [1.1612, 7.7000, 1.0027, 3.2628, 0.2719, 1.3780, 0, 0, 0.0025], ""),
    "Cu0" => ([0.0909, 34.9838, 0.4088, 11.4432, 0.5128, 3.8248, 0, 0, -0.0124],  [1.9182, 14.4904, 1.3329, 4.7301, 0.3842, 1.6394, 0, 0, 0.0035], ""),
    "Cu1" => ([0.0749, 34.9656, 0.4147, 11.7642, 0.5238, 3.8497, 0, 0, -0.0127],  [1.8814, 13.4333, 1.2809, 4.5446, 0.3646, 1.6022, 0, 0, 0.0033], ""),
    "Cu2" => ([0.0232, 34.9686, 0.4023, 11.5640, 0.5882, 3.8428, 0, 0, -0.0137],  [1.5189, 10.4779, 1.1512, 3.8132, 0.2918, 1.3979, 0, 0, 0.0017], ""),
    "Cu3" => ([0.0031, 34.9074, 0.3582, 10.9138, 0.6531, 3.8279, 0, 0, -0.0147],  [1.2797, 8.4502, 1.0315, 3.2796, 0.2401, 1.2498, 0, 0, 0.0015], ""),
    "Cu4" => ([-0.0132, 30.6817, 0.2801, 11.1626, 0.7490, 3.8172, 0, 0, -0.0165], [0.9568, 7.4481, 0.9099, 3.3964, 0.3729, 1.4936, 0, 0, 0.0049], ""),

    # 4d electrons of transition atoms and ions (P. J. Brown)
    "Y0" =>  ([0.5915, 67.6081, 1.5123, 17.9004, -1.1130, 14.1359, 0, 0, 0.0080], [14.4084, 44.6577, 5.1045, 14.9043, -0.0535, 3.3189, 0, 0, 0.0028], ""),
    "Zr0" => ([0.4106, 59.9961, 1.0543, 18.6476, -0.4751, 10.5400, 0, 0, 0.0106], [10.1378, 35.3372, 4.7734, 12.5453, -0.0489, 2.6721, 0, 0, 0.0036], ""),
    "Zr1" => ([0.4532, 59.5948, 0.7834, 21.4357, -0.2451, 9.0360, 0, 0, 0.0098],  [11.8722, 34.9200, 4.0502, 12.1266, -0.0632, 2.8278, 0, 0, 0.0034], ""),
    "Nb0" => ([0.3946, 49.2297, 1.3197, 14.8216, -0.7269, 9.6156, 0, 0, 0.0129],  [7.4796, 33.1789, 5.0884, 11.5708, -0.0281, 1.5635, 0, 0, 0.0047], ""),
    "Nb1" => ([0.4572, 49.9182, 1.0274, 15.7256, -0.4962, 9.1573, 0, 0, 0.0118],  [8.7735, 33.2848, 4.6556, 11.6046, -0.0268, 1.5389, 0, 0, 0.0044], ""),
    "Mo0" => ([0.1806, 49.0568, 1.2306, 14.7859, -0.4268, 6.9866, 0, 0, 0.0171],  [5.1180, 23.4217, 4.1809, 9.2080, -0.0505, 1.7434, 0, 0, 0.0053], ""),
    "Mo1" => ([0.3500, 48.0354, 1.0305, 15.0604, -0.3929, 7.4790, 0, 0, 0.0139],  [7.2367, 28.1282, 4.0705, 9.9228, -0.0317, 1.4552, 0, 0, 0.0049], ""),
    "Tc0" => ([0.1298, 49.6611, 1.1656, 14.1307, -0.3134, 5.5129, 0, 0, 0.0195],  [4.2441, 21.3974, 3.9439, 8.3753, -0.0371, 1.1870, 0, 0, 0.0066], ""),
    "Tc1" => ([0.2674, 48.9566, 0.9569, 15.1413, -0.2387, 5.4578, 0, 0, 0.0160],  [6.4056, 24.8243, 3.5400, 8.6112, -0.0366, 1.4846, 0, 0, 0.0044], ""),
    "Ru0" => ([0.1069, 49.4238, 1.1912, 12.7417, -0.3176, 4.9125, 0, 0, 0.0213],  [3.7445, 18.6128, 3.4749, 7.4201, -0.0363, 1.0068, 0, 0, 0.0073], ""),
    "Ru1" => ([0.4410, 33.3086, 1.4775, 9.5531, -0.9361, 6.7220, 0, 0, 0.0176],   [5.2826, 23.6832, 3.5813, 8.1521, -0.0257, 0.4255, 0, 0, 0.0131], ""),
    "Rh0" => ([0.0976, 49.8825, 1.1601, 11.8307, -0.2789, 4.1266, 0, 0, 0.0234],  [3.3651, 17.3444, 3.2121, 6.8041, -0.0350, 0.5031, 0, 0, 0.0146], ""),
    "Rh1" => ([0.3342, 29.7564, 1.2209, 9.4384, -0.5755, 5.3320, 0, 0, 0.0210],   [4.0260, 18.9497, 3.1663, 6.9998, -0.0296, 0.4862, 0, 0, 0.0127], ""),
    "Pd0" => ([0.2003, 29.3633, 1.1446, 9.5993, -0.3689, 4.0423, 0, 0, 0.0251],   [3.3105, 14.7265, 2.6332, 5.8618, -0.0437, 1.1303, 0, 0, 0.0053], ""),
    "Pd1" => ([0.5033, 24.5037, 1.9982, 6.9082, -1.5240, 5.5133, 0, 0, 0.0213],   [4.2749, 17.9002, 2.7021, 6.3541, -0.0258, 0.6999, 0, 0, 0.0071], ""),
    
    # 5d electrons of rare earth ions (P. J. Brown)
    "Ce2" => ([0.2953, 17.6846, 0.2923, 6.7329, 0.4313, 5.3827, 0, 0, -0.0194],   [0.9809, 18.0630, 1.8413, 7.7688, 0.9905, 2.8452, 0, 0, 0.0120], ""),
    "Nd2" => ([0.1645, 25.0453, 0.2522, 11.9782, 0.6012, 4.9461, 0, 0, -0.0180],  [1.4530, 18.3398, 1.6196, 7.2854, 0.8752, 2.6224, 0, 0, 0.0126], ""),
    "Nd3" => ([0.0540, 25.0293, 0.3101, 12.1020, 0.6575, 4.7223, 0, 0, -0.0216],  [0.6751, 18.3421, 1.6272, 7.2600, 0.9644, 2.6016, 0, 0, 0.0150], ""),
    "Sm2" => ([0.0909, 25.2032, 0.3037, 11.8562, 0.6250, 4.2366, 0, 0, -0.0200],  [1.0360, 18.4249, 1.4769, 7.0321, 0.8810, 2.4367, 0, 0, 0.0152], ""),
    "Sm3" => ([0.0288, 25.2068, 0.2973, 11.8311, 0.6954, 4.2117, 0, 0, -0.0213],  [0.4707, 18.4301, 1.4261, 7.0336, 0.9574, 2.4387, 0, 0, 0.0182], ""),
    "Eu2" => ([0.0755, 25.2960, 0.3001, 11.5993, 0.6438, 4.0252, 0, 0, -0.0196],  [0.8970, 18.4429, 1.3769, 7.0054, 0.9060, 2.4213, 0, 0, 0.0190], ""),
    "Eu3" => ([0.0204, 25.3078, 0.3010, 11.4744, 0.7005, 3.9420, 0, 0, -0.0220],  [0.3985, 18.4514, 1.3307, 6.9556, 0.9603, 2.3780, 0, 0, 0.0197], ""),
    "Gd2" => ([0.0636, 25.3823, 0.3033, 11.2125, 0.6528, 3.7877, 0, 0, -0.0199],  [0.7756, 18.4695, 1.3124, 6.8990, 0.8956, 2.3383, 0, 0, 0.0199], ""),
    "Gd3" => ([0.0186, 25.3867, 0.2895, 11.1421, 0.7135, 3.7520, 0, 0, -0.0217],  [0.3347, 18.4758, 1.2465, 6.8767, 0.9537, 2.3184, 0, 0, 0.0217], ""),
    "Tb2" => ([0.0547, 25.5086, 0.3171, 10.5911, 0.6490, 3.5171, 0, 0, -0.0212],  [0.6688, 18.4909, 1.2487, 6.8219, 0.8888, 2.2751, 0, 0, 0.0215], ""),
    "Tb3" => ([0.0177, 25.5095, 0.2921, 10.5769, 0.7133, 3.5122, 0, 0, -0.0231],  [0.2892, 18.4973, 1.1678, 6.7972, 0.9437, 2.2573, 0, 0, 0.0232], ""),
    "Dy2" => ([0.1308, 18.3155, 0.3118, 7.6645, 0.5795, 3.1469, 0, 0, -0.0226],   [0.5917, 18.5114, 1.1828, 6.7465, 0.8801, 2.2141, 0, 0, 0.0229], ""),
    "Dy3" => ([0.1157, 15.0732, 0.3270, 6.7991, 0.5821, 3.0202, 0, 0, -0.0249],   [0.2523, 18.5172, 1.0914, 6.7362, 0.9345, 2.2082, 0, 0, 0.0250], ""),
    "Ho2" => ([0.0995, 18.1761, 0.3305, 7.8556, 0.5921, 2.9799, 0, 0, -0.0230],   [0.5094, 18.5155, 1.1234, 6.7060, 0.8727, 2.1589, 0, 0, 0.0242], ""),
    "Ho3" => ([0.0566, 18.3176, 0.3365, 7.6880, 0.6317, 2.9427, 0, 0, -0.0248],   [0.2188, 18.5157, 1.0240, 6.7070, 0.9251, 2.1614, 0, 0, 0.0268], ""),
    "Er2" => ([0.1122, 18.1223, 0.3462, 6.9106, 0.5649, 2.7614, 0, 0, -0.0235],   [0.4693, 18.5278, 1.0545, 6.6493, 0.8679, 2.1201, 0, 0, 0.0261], ""),
    "Er3" => ([0.0586, 17.9802, 0.3540, 7.0964, 0.6126, 2.7482, 0, 0, -0.0251],   [0.1710, 18.5337, 0.9879, 6.6246, 0.9044, 2.1004, 0, 0, 0.0278], ""),
    "Tm2" => ([0.0983, 18.3236, 0.3380, 6.9178, 0.5875, 2.6622, 0, 0, -0.0241],   [0.4198, 18.5417, 0.9959, 6.6002, 0.8593, 2.0818, 0, 0, 0.0284], ""),
    "Tm3" => ([0.0581, 15.0922, 0.2787, 7.8015, 0.6854, 2.7931, 0, 0, -0.0224],   [0.1760, 18.5417, 0.9105, 6.5787, 0.8970, 2.0622, 0, 0, 0.0294], ""),
    "Yb2" => ([0.0855, 18.5123, 0.2943, 7.3734, 0.6412, 2.6777, 0, 0, -0.0213],   [0.3852, 18.5497, 0.9415, 6.5507, 0.8492, 2.0425, 0, 0, 0.0301], ""),
    "Yb3" => ([0.0416, 16.0949, 0.2849, 7.8341, 0.6961, 2.6725, 0, 0, -0.0229],   [0.1570, 18.5553, 0.8484, 6.5403, 0.8880, 2.0367, 0, 0, 0.0318], ""),
    "Pr3" => ([0.0504, 24.9989, 0.2572, 12.0377, 0.7142, 5.0039, 0, 0, -0.0219],  [0.8734, 18.9876, 1.5594, 6.0872, 0.8142, 2.4150, 0, 0, 0.0111], ""),

    # 5d electrons of actinide ions (P. J. Brown)
    "U3" =>  ([0.5058, 23.2882, 1.3464, 7.0028, -0.8724, 4.8683, 0, 0, 0.0192],   [4.1582, 16.5336, 2.4675, 5.9516, -0.0252, 0.7646, 0, 0, 0.0057], ""),
    "U4" =>  ([0.3291, 23.5475, 1.0836, 8.4540, -0.4340, 4.1196, 0, 0, 0.0214],   [3.7449, 13.8944, 2.6453, 4.8634, -0.5218, 3.1919, 0, 0, 0.0009], ""),
    "U5" =>  ([0.3650, 19.8038, 3.2199, 6.2818, -2.6077, 5.3010, 0, 0, 0.0233],   [3.0724, 12.5460, 2.3076, 5.2314, -0.0644, 1.4738, 0, 0, 0.0035], ""),
    "Np3" => ([0.5157, 20.8654, 2.2784, 5.8930, -1.8163, 4.8457, 0, 0, 0.0211],   [3.7170, 15.1333, 2.3216, 5.5025, -0.0275, 0.7996, 0, 0, 0.0052], ""),
    "Np4" => ([0.4206, 19.8046, 2.8004, 5.9783, -2.2436, 4.9848, 0, 0, 0.0228],   [2.9203, 14.6463, 2.5979, 5.5592, -0.0301, 0.3669, 0, 0, 0.0141], ""),
    "Np5" => ([0.3692, 18.1900, 3.1510, 5.8500, -2.5446, 4.9164, 0, 0, 0.0248],   [2.3308, 13.6540, 2.7219, 5.4935, -0.1357, 0.0493, 0, 0, 0.1224], ""),
    "Np6" => ([0.2929, 17.5611, 3.4866, 5.7847, -2.8066, 4.8707, 0, 0, 0.0267],   [1.8245, 13.1803, 2.8508, 5.4068, -0.1579, 0.0444, 0, 0, 0.1438], ""),
    "Pu3" => ([0.3840, 16.6793, 3.1049, 5.4210, -2.5148, 4.5512, 0, 0, 0.0263],   [2.0885, 12.8712, 2.5961, 5.1896, -0.1465, 0.0393, 0, 0, 0.1343], ""),
    "Pu4" => ([0.4934, 16.8355, 1.6394, 5.6384, -1.1581, 4.1399, 0, 0, 0.0248],   [2.7244, 12.9262, 2.3387, 5.1633, -0.1300, 0.0457, 0, 0, 0.1177], ""),
    "Pu5" => ([0.3888, 16.5592, 2.0362, 5.6567, -1.4515, 4.2552, 0, 0, 0.0267],   [2.1409, 12.8319, 2.5664, 5.1522, -0.1338, 0.0457, 0, 0, 0.1210], ""),
    "Pu6" => ([0.3172, 16.0507, 3.4654, 5.3507, -2.8102, 4.5133, 0, 0, 0.0281],   [1.7262, 12.3240, 2.6652, 5.0662, -0.1695, 0.0406, 0, 0, 0.1550], ""),
    "Am2" => ([0.4743, 21.7761, 1.5800, 5.6902, -1.0779, 4.1451, 0, 0, 0.0218],   [3.5237, 15.9545, 2.2855, 5.1946, -0.0142, 0.5853, 0, 0, 0.0033], ""),
    "Am3" => ([0.4239, 19.5739, 1.4573, 5.8722, -0.9052, 3.9682, 0, 0, 0.0238],   [2.8622, 14.7328, 2.4099, 5.1439, -0.1326, 0.0309, 0, 0, 0.1233], ""),
    "Am4" => ([0.3737, 17.8625, 1.3521, 6.0426, -0.7514, 3.7199, 0, 0, 0.0258],   [2.4141, 12.9478, 2.3687, 4.9447, -0.2490, 0.0215, 0, 0, 0.2371], ""),
    "Am5" => ([0.2956, 17.3725, 1.4525, 6.0734, -0.7755, 3.6619, 0, 0, 0.0277],   [2.0109, 12.0534, 2.4155, 4.8358, -0.2264, 0.0275, 0, 0, 0.2128], ""),
    "Am6" => ([0.2302, 16.9533, 1.4864, 6.1159, -0.7457, 3.5426, 0, 0, 0.0294],   [1.6778, 11.3372, 2.4531, 4.7247, -0.2043, 0.0337, 0, 0, 0.1892], ""),
    "Am7" => ([0.3601, 12.7299, 1.9640, 5.1203, -1.3560, 3.7142, 0, 0, 0.0316],   [1.8845, 9.1606, 2.0746, 4.0422, -0.1318, 1.7227, 0, 0, 0.0020], ""),

    # 5d electrons of transition atoms and ions from Hf to Re. (Kobayashi, Nagao, Ito)
    "Hf2"  => ([0.4229, 50.465, 0.7333, 23.865, -0.3798, 4.051, 0.2252, 2.497, -0.0018],   [9.6670, 33.435, 5.2429, 13.529, -0.5533, 1.402, 0.4934, 1.254, -0.0033],  "6s⁰5d²"),
    "Hf3"  => ([0.3555, 40.954, 0.8483, 21.726, -0.4116, 4.305, 0.2101, 2.349, -0.0023],   [7.5646, 27.367, 5.0743, 12.402, -0.4133, 1.742, 0.3163, 1.437, -0.0012],  "6s⁰5d¹"),
    "Ta2"  => ([0.3976, 45.095, 0.7746, 21.028, -0.6098, 3.471, 0.4395, 2.570, -0.0020],   [8.1746, 29.871, 4.9405, 12.188, -1.1294, 1.254, 1.0658, 1.181, -0.0046],  "6s⁰5d³"),
    "Ta3"  => ([0.3611, 36.921, 0.8579, 19.195, -0.4945, 3.857, 0.2781, 2.303, -0.0026],   [6.5998, 25.026, 4.7913, 11.303, -0.7801, 1.529, 0.6820, 1.382, -0.0019],  "6s⁰5d²"),
    "Ta4"  => ([0.3065, 31.817, 0.9611, 17.749, -0.5463, 3.979, 0.2816, 2.232, -0.0030],   [5.5048, 21.620, 4.6734, 10.556, -0.2181, 2.100, 0.0810, 1.049, -0.0016],  "6s⁰5d¹"),
    "W0a"  => ([0.3990, 73.810, 0.7138, 22.815, -2.0436, 2.710, 1.9319, 2.559, -0.0023],   [8.4216, 64.652, 7.6983, 23.124, 2.3605, 8.583, -0.3876, 4.620, 0.0033],   "6s⁰5d⁶"),
    "W0b"  => ([0.3811, 62.707, 0.7523, 21.434, -12.5449, 2.702, 12.4130, 2.674, -0.0023], [6.9289, 53.425, 7.2157, 20.016, 5.7754, 6.543, -4.2114, 5.854, 0.0036],   "6s¹5d⁵"),
    "W0c"  => ([0.3653, 53.965, 0.7926, 20.078, -0.8142, 3.030, 0.6581, 2.476, -0.0023],   [5.4017, 47.110, 6.9113, 18.760, 7.0586, 6.363, -5.4525, 5.802, 0.0040],   "6s²5d⁴"),
    "W1a"  => ([0.4077, 51.367, 0.7436, 20.256, -9.8283, 2.780, 9.6788, 2.740, -0.0021],   [5.6854, 44.743, 6.8903, 18.753, 7.8523, 6.308, -6.3447, 5.842, 0.0038],   "6s⁰5d⁵"),
    "W1b"  => ([0.3834, 46.233, 0.7890, 19.278, -1.4650, 2.947, 1.2945, 2.628, -0.0022],   [4.5118, 41.281, 6.7136, 17.964, 7.6331, 6.269, -6.0365, 5.758, 0.0042],   "6s¹5d⁴"),
    "W2c"  => ([0.3876, 40.340, 0.8008, 18.621, -1.3911, 2.995, 1.2048, 2.627, -0.0023],   [7.0301, 26.990, 4.6604, 11.083, -0.8220, 1.145, 0.7581, 1.042, -0.0068],  "6s⁰5d⁴"),
    "W3"   => ([0.3610, 33.519, 0.8717, 17.176, -0.6183, 3.445, 0.3883, 2.276, -0.0028],   [5.8036, 22.969, 4.5243, 10.361, -0.7897, 1.388, 0.6927, 1.248, -0.0031],  "6s⁰5d³"),
    "W4"   => ([0.3221, 29.047, 0.9574, 15.979, -0.6287, 3.597, 0.3525, 2.174, -0.0033],   [4.9006, 20.117, 4.4360, 9.765, -2.0009, 1.497, 1.8714, 1.431, -0.0014],   "6s⁰5d²"),
    "W5"   => ([0.2725, 25.966, 1.0558, 14.954, -0.6990, 3.643, 0.3745, 2.145, -0.0037],   [4.1973, 17.967, 4.3791, 9.255, -1.8830, 1.603, 1.7205, 1.518, -0.0004],   "6s⁰5d¹"),
    "Re0a" => ([0.3807, 63.042, 0.7497, 19.967, -6.5300, 2.501, 6.4013, 2.451, -0.0028],   [6.7574, 55.529, 6.7931, 20.125, 2.3113, 7.529, -0.5004, 4.412, 0.0037],   "6s⁰5d⁷"),
    "Re0b" => ([0.3691, 53.934, 0.7837, 18.790, -9.1491, 2.558, 8.9983, 2.517, -0.0027],   [5.5830, 46.852, 6.4516, 17.855, 5.0609, 6.037, -3.5427, 5.312, 0.0040],   "6s¹5d⁶"),
    "Re0c" => ([0.3548, 47.108, 0.8210, 17.769, -9.8674, 2.599, 9.6938, 2.556, -0.0027],   [4.4322, 41.798, 6.2063, 16.844, 9.8763, 5.751, -8.3294, 5.406, 0.0043],   "6s²5d⁵"),
    "Re1a" => ([0.3944, 45.427, 0.7742, 17.948, -3.1692, 2.653, 3.0028, 2.521, -0.0026],   [4.7231, 40.130, 6.2139, 16.902, 8.8240, 5.766, -7.3608, 5.396, 0.0042],   "6s⁰5d⁶"),
    "Re1b" => ([0.3736, 41.151, 0.8160, 17.158, -7.0396, 2.642, 6.8523, 2.577, -0.0026],   [3.7875, 37.254, 6.0587, 16.264, 6.9896, 5.786, -5.4457, 5.279, 0.0045],   "6s¹5d⁵"),
    "Re2"  => ([0.3825, 36.336, 0.8218, 16.636, -8.7220, 2.657, 8.5201, 2.601, -0.0026],   [3.0708, 33.896, 5.9941, 15.817, 6.7816, 5.768, -5.2056, 5.231, 0.0047],   "6s⁰5d⁵"),
    "Re3"  => ([0.3585, 30.671, 0.8863, 15.527, -0.8682, 3.047, 0.6263, 2.280, -0.0030],   [1.7870, 31.248, 5.9068, 15.160, 17.4262, 5.551, -15.6856, 5.339, 0.0053], "6s⁰5d⁴"),
    "Re4"  => ([0.2974, 27.372, 0.9826, 14.807, -1.8869, 2.840, 1.6100, 2.476, -0.0031],   [4.4033, 18.733, 4.1971, 9.047, -2.3106, 1.352, 2.1866, 1.299, -0.0025],   "6s⁰5d³"),
    "Re5"  => ([0.3143, 23.522, 1.0276, 13.505, -0.7438, 3.393, 0.4059, 2.030, -0.0041],   [3.8162, 16.843, 4.1400, 8.605, -4.0882, 1.443, 3.9333, 1.408, -0.0012],   "6s⁰5d²"),
    "Re6"  => ([0.2146, 22.496, 1.1616, 13.064, -1.0455, 3.162, 0.6734, 2.196, -0.0041],   [3.3349, 15.341, 4.1015, 8.213, -2.3511, 1.531, 2.1651, 1.459, -0.0003],   "6s⁰5d¹"),

    # 5d electrons of transition atoms and ions from Os to Au. (Kobayashi, Nagao, Ito)
    "Os0a" => ([0.3676, 54.835, 0.7793, 17.716, -2.0669, 2.418, 1.9224, 2.247, -0.0034],   [5.5418, 48.893, 6.0803, 17.984, 2.2542, 6.853, -0.5285, 4.095, 0.0040],   "6s⁰5d⁸"),
    "Os0b" => ([0.3571, 47.458, 0.8123, 16.770, -1.2072, 2.556, 1.0404, 2.211, -0.0033],   [4.6511, 41.610, 5.8194, 16.062, 6.7172, 5.447, -5.2689, 4.986, 0.0043],   "6s¹5d⁷"),
    "Os0c" => ([0.3467, 41.778, 0.8458, 15.918, -5.6370, 2.459, 5.4472, 2.381, -0.0032],   [3.7421, 37.491, 5.6137, 15.256, 10.0488, 5.293, -8.5683, 4.991, 0.0046],  "6s²5d⁶"),
    "Os1a" => ([0.3837, 40.665, 0.8006, 16.096, -3.5305, 2.487, 3.3488, 2.366, -0.0030],   [4.0236, 36.272, 5.6349, 15.338, 8.6568, 5.318, -7.2505, 4.980, 0.0044],   "6s⁰5d⁷"),
    "Os1b" => ([0.3666, 36.997, 0.8390, 15.425, -2.6944, 2.537, 2.4916, 2.360, -0.0031],   [3.2611, 33.818, 5.4945, 14.808, 5.7610, 5.391, -4.2811, 4.827, 0.0048],   "6s¹5d⁶"),
    "Os2"  => ([0.3786, 33.005, 0.8412, 14.990, -7.0632, 2.503, 6.8462, 2.433, -0.0030],   [2.7289, 30.900, 5.4357, 14.396, 18.6546, 5.151, -17.1708, 4.993, 0.0050], "6s⁰5d⁶"),
    "Os3"  => ([0.3557, 28.222, 0.9002, 14.140, -2.5972, 2.601, 2.3444, 2.376, -0.0032],   [4.5913, 19.692, 4.0615, 8.862, -0.9950, 1.086, 0.9081, 0.986, -0.0077],   "6s⁰5d⁵"),
    "Os4"  => ([0.3337, 24.723, 0.9655, 13.288, -0.9653, 2.906, 0.6698, 2.117, -0.0037],   [3.9724, 17.514, 3.9817, 8.421, -0.7719, 1.264, 0.6552, 1.107, -0.0042],   "6s⁰5d⁴"),
    "Os5"  => ([0.3055, 22.152, 1.0395, 12.529, -0.9158, 3.016, 0.5750, 2.032, -0.0042],   [3.4764, 15.826, 3.9241, 8.032, -0.6854, 1.397, 0.5394, 1.184, -0.0024],   "6s⁰5d³"),
    "Os6"  => ([0.2714, 20.218, 1.1211, 11.851, -0.9773, 3.050, 0.5894, 2.005, -0.0046],   [3.0642, 14.479, 3.8847, 7.686, -4.7235, 1.384, 4.5485, 1.352, -0.0012],   "6s⁰5d²"),
    "Os7"  => ([0.2101, 19.108, 1.2240, 11.347, -1.2543, 2.933, 0.8250, 2.088, -0.0048],   [2.7164, 13.366, 3.8554, 7.370, -5.0211, 1.439, 4.8166, 1.405, -0.0004],   "6s⁰5d¹"),
    "Ir0a" => ([0.3564, 48.464, 0.8049, 15.923, -2.5258, 2.265, 2.3675, 2.121, -0.0040],   [4.6102, 43.878, 5.4892, 16.394, 2.1707, 6.423, -0.4761, 3.722, 0.0043],   "6s⁰5d⁹"),
    "Ir0b" => ([0.3492, 42.195, 0.8350, 15.113, -5.1496, 2.279, 4.9686, 2.201, -0.0038],   [3.9372, 37.508, 5.2846, 14.661, 3.7267, 5.263, -2.3158, 4.416, 0.0045],   "6s¹5d⁸"),
    "Ir0c" => ([0.3400, 37.499, 0.8675, 14.402, -2.3703, 2.370, 2.1661, 2.177, -0.0037],   [3.2263, 33.922, 5.1086, 13.921, 6.5993, 4.978, -5.1841, 4.549, 0.0048],   "6s²5d⁷"),
    "Ir1a" => ([0.3744, 36.764, 0.8240, 14.576, -8.8616, 2.303, 8.6664, 2.255, -0.0035],   [3.4956, 32.991, 5.1369, 13.998, 8.3991, 4.924, -7.0561, 4.612, 0.0047],   "6s⁰5d⁸"),
    "Ir1b" => ([0.3604, 33.570, 0.8597, 13.993, -2.1686, 2.412, 1.9518, 2.188, -0.0036],   [2.8732, 30.809, 5.0094, 13.522, 6.8656, 4.933, -5.4669, 4.526, 0.0050],   "6s¹5d⁷"),
    "Ir2"  => ([0.3802, 30.032, 0.8550, 13.567, -1.6185, 2.488, 1.3866, 2.162, -0.0035],   [2.4419, 28.356, 4.9470, 13.222, 4.7478, 5.029, -3.3259, 4.393, 0.0052],   "6s⁰5d⁷"),
    "Ir3"  => ([0.3678, 25.828, 0.9065, 12.788, -0.8587, 2.745, 0.5883, 1.960, -0.0040],   [1.5883, 25.969, 4.8472, 12.711, 5.6507, 4.949, -4.1190, 4.388, 0.0056],   "6s⁰5d⁶"),
    "Ir4"  => ([0.3969, 22.050, 0.9310, 11.768, -0.7090, 3.017, 0.3857, 1.778, -0.0047],   [3.5964, 16.439, 3.7872, 7.873, -1.2303, 1.092, 1.1232, 1.003, -0.0068],   "6s⁰5d⁵"),
    "Ir5"  => ([0.3059, 20.759, 1.0432, 11.622, -1.3457, 2.655, 1.0008, 2.066, -0.0043],   [3.1710, 14.923, 3.7320, 7.529, -1.5400, 1.200, 1.4048, 1.119, -0.0040],   "6s⁰5d⁴"),
    "Ir6"  => ([0.2854, 18.867, 1.1123, 10.982, -1.1403, 2.782, 0.7475, 1.959, -0.0049],   [2.8163, 13.697, 3.6906, 7.219, -5.2957, 1.254, 5.1328, 1.228, -0.0024],   "6s⁰5d³"),
    "Pt1"  => ([0.3664, 33.503, 0.8449, 13.303, -2.6192, 2.227, 2.4116, 2.059, -0.0041],   [3.0783, 30.198, 4.7065, 12.857, 5.7270, 4.650, -4.4425, 4.222, 0.0048],   "6s⁰5d⁹"),
    "Pt2"  => ([0.3701, 27.799, 0.8761, 12.480, -3.2959, 2.258, 3.0535, 2.106, -0.0040],   [2.2407, 26.003, 4.5300, 12.118, 6.7189, 4.557, -5.3954, 4.187, 0.0053],   "6s⁰5d⁸"),
    "Pt3"  => ([0.3606, 24.052, 0.9205, 11.833, -4.0612, 2.271, 3.7841, 2.133, -0.0041],   [1.5391, 23.689, 4.4263, 11.645, 8.5673, 4.482, -7.1591, 4.181, 0.0057],   "6s⁰5d⁷"),
    "Pt4"  => ([0.3429, 21.358, 0.9753, 11.261, -2.2859, 2.357, 1.9719, 2.074, -0.0043],   [0.9477, 22.775, 4.3337, 11.319, 7.2374, 4.500, -5.7086, 4.106, 0.0062],   "6s⁰5d⁶"),
    "Pt5"  => ([0.3649, 18.762, 1.0102, 10.456, -0.8785, 2.765, 0.5087, 1.737, -0.0053],   [0.5132, 23.124, 4.1835, 11.009, 7.4312, 4.478, -5.7759, 4.065, 0.0066],   "6s⁰5d⁵"),
    "Pt6"  => ([0.2877, 17.780, 1.1111, 10.263, -1.9416, 2.432, 1.5477, 2.015, -0.0049],   [2.5907, 12.991, 3.5155, 6.801, -0.7511, 1.206, 0.6014, 1.022, -0.0043],   "6s⁰5d⁴"),
    "Au1"  => ([0.3475, 28.294, 0.8964, 11.794, -4.5383, 2.086, 4.2988, 1.983, -0.0047],   [2.3189, 25.954, 4.2267, 11.444, 9.0997, 4.215, -7.8523, 3.981, 0.0052],   "6s¹5d⁹"),
    "Au2"  => ([0.3664, 25.714, 0.8919, 11.488, -1.7302, 2.207, 1.4763, 1.908, -0.0046],   [2.0676, 23.960, 4.1678, 11.164, 13.3827, 4.151, -12.1447, 3.996, 0.0054], "6s⁰5d⁹"),
    "Au3"  => ([0.3639, 22.330, 0.9313, 10.883, -1.1359, 2.355, 0.8455, 1.809, -0.0048],   [1.4776, 21.784, 4.0585, 10.746, 5.9059, 4.242, -4.5905, 3.842, 0.0058],   "6s⁰5d⁸"),
    "Au4"  => ([0.3472, 19.944, 0.9807, 10.424, -3.8228, 2.169, 3.4995, 2.009, -0.0047],   [2.9775, 14.632, 3.4472, 6.956, -1.9784, 0.794, 1.8980, 0.745, -0.0182],   "6s⁰5d⁷"),
    "Au5"  => ([0.3292, 18.073, 1.0376, 9.957, -2.2028, 2.257, 1.8410, 1.938, -0.0050],    [0.5763, 20.517, 3.8572, 10.171, 6.0222, 4.213, -4.5020, 3.760, 0.0067],   "6s⁰5d⁶"),
)
